/*****************************************************************************
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: ezplot.cpp,v 1.1 2000/06/19 18:05:03 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/
/*----------------------------------------------------------------------*/
/*				EZPLOT					*/
/*----------------------------------------------------------------------*/

#include "ezplot.h"
#include "algo.h"

static int plot (SGP_ID *gid);
static void drawaxis(void);
static void symbol (int sym, double symwidth, double symheight);


/*-------------------------------------*/
/* GLOBAL variables for EZPLOT & EZSET */
/*-------------------------------------*/

bool ezplot_firstcall = TRUE;	/* set to false on first call to EZSET or EZPLOT */
struct ezplot_var ez;		/* ezplot variables */

/*----------------*/
/* DEFAULT values */
/*----------------*/

#define TICKRATIO 0.4		/* ratio of minor to major tick lengths */
#define MAXNUMFMT 15		/* maximum length of a numeric format */

#define DEF_CRTMODE    0  /*GM_HIGHRES*/	/* default crt mode */
#define DEF_CHARHEIGHT (1./43.)	/* default size of characters in NDC */
#define DEF_CHARWIDTH  (1./80.)

#define DEF_CURVE_CLR   C_YELLOW


/*----------------------------------------------------------------------*/
/* 				GLOBAL VARIABLES			*/
/*----------------------------------------------------------------------*/

static
double charwidth, charheight;	/* Size of characters in NDC */

static
double  xp_min, xp_max,
    yp_min, yp_max;		/* boundry of plot frame in NDC */
static
double  xa_min, xa_max,
    ya_min, ya_max;		/* extent of axes in NDC */
static
double  xgw_min, xgw_max,
    ygw_min, ygw_max;	/* boundary of graph in input coords */
static
double  xgn_min, xgn_max,
    ygn_min, ygn_max;	/* boundy of graph in NDC */
static
double xt_min, xt_max,
    yt_min, yt_max;		/* boundary of axis ticks */

static
double  xl_min, xl_max,
    yl_min, yl_max;		/* boundary of legend box */

static  double title_row;	/* y-coord of title row */

static  double xtl_ofs;		/* Offset y-coord of x tick labels from axis */
static	double ytl_ofs;		/* Offset x-coord of y tick labels from axis */

static	double xlbl_row;	/* row of x label in world coord */
static	double ylbl_col;	/* column of y label in world coord */

static
double xw_tickinc, yw_tickinc;	/* increment between major ticks in WC */

static
double xn_tickinc, yn_tickinc;	/* increment between major ticks in NDC */

static int x_nint, y_nint;	/* number of intervals along x & y axes */
static int x_fldwid, x_frac;	/* numeric field sizes & number of digits   */
static int y_fldwid, y_frac;	/* in fraction of number, used for printf() */

static double xtl_wid, ytl_wid;	/* length of ticks labels in NDC */
static double tl_height;	/* height of tick labels in NDC */

static char x_numfmt[10];	/* format to print x tick labels */
static char y_numfmt[10];	/* format to print y tick labels */


/*----------------------------------------------------------------------
 * FUNCTION IDENTICIFATION
 *
 *	Name:		EZPLOT
 *	Programmer:	Kevin Rosenberg
 *	Date:		 5-83 Version 1
 *			11-84 Version 2
 *
 * SYNOPSIS
 *	ezplot (x, y, n)
 *	double x[]	x-coordinates of point  makeing up a curve 
 *	double y[]	y-coordinates of points makeing up a curve
 *	int    n	Number of points in curve
 *
 * DESCRIPTION
 *	    This is a sophisticated plotting program.
 *
 *----------------------------------------------------------------------*/


SGP_ID
ezplot (double x[], double y[], int num)
{
    unsigned int size, i;
    SGP_ID gid = NULL;

    if (ez.i_plotimmediate == TRUE) {
	plot (&gid);
	return (gid);
    }

    if (num < 1)
	return (NULL);

    if (ezplot_firstcall == TRUE)
	ezinit();

    if (ez.o_unknowncurves == FALSE && ez.i_numcurves >= ez.o_reqcurves)
	ezclear ();

    if (ez.i_numcurves < MAXCURVES &&
	(ez.o_unknowncurves == TRUE || ez.i_numcurves < ez.o_reqcurves)) {
	size = num * sizeof(double);
	ez.curve[ez.i_numcurves].x = new double [size];
	ez.curve[ez.i_numcurves].y = new double [size];
	for (i = 0; i < num; i++) {
	    ez.curve[ez.i_numcurves].x[i] = x[i];
	    ez.curve[ez.i_numcurves].y[i] = y[i];
	}
	ez.curve[ez.i_numcurves].numpts = num;
	ez.curve[ez.i_numcurves].color = ez.o_color;
	ez.curve[ez.i_numcurves].linestyle = ez.o_linestyle;
	ez.curve[ez.i_numcurves].symbol = ez.o_symbol;
	ez.curve[ez.i_numcurves].symfreq = ez.o_symfreq;
	strncpy (ez.curve[ez.i_numcurves].legend, ez.c_legend, MAXLEGEND);
	++ez.i_numcurves;
    }
    
    if (ez.o_unknowncurves == FALSE && ez.i_numcurves >= ez.o_reqcurves) {
	plot (&gid);
	ezclear ();
    }

    return (gid);
}


/* NAME
 *   plot		INTERNAL: Plots all curves collected by ezplot()
 *
 * SYNOPSIS
 *   plot()
 *
 * DESCRIPTION
 *   This routine plots the curves that have stored by ezplot().
 *
 * CALLED BY
 *   EZPLOT() only
 *
 * CALLS
 *   drawaxis() & make_numfmt()
 */

static int plot (SGP_ID *gid)
{
    double xmin, xmax;		/* extend of curves in world coord  */
    double ymin, ymax;
    double x_added_ticks;		/* number of tick spaces added to axis */
    double y_added_ticks;
    double symwidth, symheight;	/* size of symbol in NDC */
    double leg_width, leg_height;	/* size of legend box */
    int num_leg;			/* number of legend titles */
    int max_leg;			/* longest legend in characters */
    int i, j, ip, ic, n;

    if (ez.i_numcurves <= 0)
	return (FALSE);

    if (ez.o_ustart == FALSE && *gid == NULL) {
	*gid = sgp2_init (0, 0, "Ezplot");
    }
    
    if (ez.s_textsize == TRUE)
	charheight = ez.v_textsize;
    else
	charheight = DEF_CHARHEIGHT;
    
    gp_set_aspect (CRTDEV, 1.0);
    
    xmin = xmax = ez.curve[0].x[0];
    ymin = ymax = ez.curve[0].y[0];
    
    for (ic = 0; ic < ez.i_numcurves; ic++) {
	for (ip = 0; ip < ez.curve[ic].numpts; ip++) {
	    if (ez.curve[ic].x[ip] > xmax)
		xmax = ez.curve[ic].x[ip];
	    else if (ez.curve[ic].x[ip] < xmin)
		xmin = ez.curve[ic].x[ip];
	    if (ez.curve[ic].y[ip] > ymax)
		ymax = ez.curve[ic].y[ip];
	    else if (ez.curve[ic].y[ip] < ymin)
		ymin = ez.curve[ic].y[ip];
	}
    }
    
    /* extend graph limits for user defined axis cross positions */
    
    if (ez.s_xcross == TRUE)
	{
	    if (ez.v_xcross < xmin)
		xmin = ez.v_xcross;
	    else if (ez.v_xcross > xmax)
		xmax = ez.v_xcross;
	}
    
    if (ez.s_ycross == TRUE)
	{
	    if (ez.v_ycross < ymin)
		ymin = ez.v_ycross;
	    else if (ez.v_ycross > ymax)
		ymax = ez.v_ycross;
	}
    
    /* find nice endpoints for axes */

    axis_scale (xmin, xmax, ez.o_xmajortick - 1, &xgw_min, &xgw_max, &x_nint);
    axis_scale (ymin, ymax, ez.o_ymajortick - 1, &ygw_min, &ygw_max, &y_nint);
    
    /* check if user set x-axis extents */
    
    if (ez.s_xmin == TRUE) {
	xgw_min = ez.v_xmin;
	x_nint = ez.o_xmajortick - 1;
    }
    if (ez.s_xmax == TRUE) {
	xgw_max = ez.v_xmax;
	x_nint = ez.o_xmajortick - 1;
    }
    
    /* check if user set y-axis extents */
    
    if (ez.s_ymin == TRUE) {
	ygw_min = ez.v_ymin;
	y_nint = ez.o_ymajortick - 1;
    }
    if (ez.s_ymax == TRUE) {
	ygw_max = ez.v_ymax;
	y_nint = ez.o_ymajortick - 1;
    }
    
    /* calculate increment between major axis in world coordinates */
    
    xw_tickinc = (xgw_max - xgw_min) / x_nint;
    yw_tickinc = (ygw_max - ygw_min) / y_nint;
    
    /* we have now calcuated xgw_min, xyw_max, ygw_min, & ygw_max */
    
    /* set the number of decimal point to users' setting or default */
    /* Two formats for numbers
       Fixed:    -nnn.f
       Exponent: -n.fffE+eee
    */
    
    if (ez.s_lxfrac == TRUE)
	x_frac = ez.v_lxfrac;
    else
	x_frac = -1;
    
    if (ez.s_lyfrac == TRUE)
	y_frac = ez.v_lyfrac;
    else
	y_frac = -1;
    
    make_numfmt (x_numfmt, &x_fldwid, &x_frac, xgw_min, xgw_max, x_nint);
    make_numfmt (y_numfmt, &y_fldwid, &y_frac, ygw_min, ygw_max, y_nint);
    
    xtl_wid = x_fldwid * charwidth;		/* calc size of tick labels */
    ytl_wid = y_fldwid * charwidth;
    tl_height = charheight;
    
    /* calculate the extent of the plot frame */
    
    xp_min = ez.o_xporigin;
    yp_min = ez.o_yporigin;
    xp_max = xp_min + ez.o_xlength;
    yp_max = yp_min + ez.o_ylength;

    xp_min = clamp (xp_min, 0., 1.);
    xp_max = clamp (xp_max, 0., 1.);
    yp_min = clamp (yp_min, 0., 1.);
    yp_max = clamp (yp_max, 0., 1.);
    
    xa_min = xp_min;		/* extent of axes */
    xa_max = xp_max;
    ya_min = yp_min;
    ya_max = yp_max;
    
    /* adjust frame for title */
    
    if (strlen(ez.c_title) != 0)
	ya_max -= 2.5 * charheight;
    title_row = ya_max + 0.5 * charheight;
    
    /* calculate legend box boundaries */
    
    max_leg = 0;			/* longest legend in characters */
    num_leg = 0;			/* number of legend titles */
    for (i = 0; i < ez.i_numcurves; i++)
	if ((n = strlen(ez.curve[i].legend)) > 0) {
	    ++num_leg;
	    max_leg = max(max_leg, n);
	}
    
    if (num_leg > 0 && ez.o_legendbox != NOLEGEND) {
	leg_width  = (max_leg + 2) * charwidth;
	leg_height = num_leg * 3 * charheight;
	
	if (ez.s_xlegend == TRUE)
	    xl_max = ez.v_xlegend;
	else {
	    xl_max = xa_max;
	    if (ez.o_legendbox == OUTSIDE)
		xa_max -= (leg_width + 0.5 * charwidth);
	}
	xl_min = xl_max - leg_width;

	if (ez.s_ylegend == TRUE)
	    yl_max = ez.v_ylegend;
	else
	    yl_max = ya_max;

	yl_min = yl_max - leg_height;

	sgp2_window  (xl_min, yl_min, xl_max, yl_max);
	sgp2_viewport (xl_min, yl_min, xl_max, yl_max);

	sgp2_color (ez.clr_legend);
	sgp2_draw_rect (xl_min, yl_min, xl_max, yl_max);

	n = 0;			/* current legend position */
	for (i = 0; i < ez.i_numcurves; i++) {
	    double xmin, xmax, xinc, y;

	    if (strlen(ez.curve[i].legend) == 0)
		continue;

	    xmin = xl_min + 1.0 * charwidth;
	    xmax = xl_max - 1.0 * charwidth;
	    y = yl_max - (2.0 + n * 3) * charheight;

	    sgp2_move_abs (xmin, y + 0.5 * charheight);
	    sgp2_draw_text (ez.curve[i].legend);
	    sgp2_color (ez.curve[i].color);
	    if (ez.curve[i].linestyle != LS_NOLINE) {
		sgp2_line_style (ez.curve[i].linestyle);
		sgp2_move_abs (xmin, y);
		sgp2_line_abs (xmax, y);
	    }
	    if (ez.curve[i].symbol > 0) {
		xinc = (xmax - xmin) / (5 - 1);
		sgp2_line_style (LS_SOLID);
		for (j = 0; j < 5; j++) {
		    sgp2_move_abs (xmin + j * xinc, y);
		    symbol(ez.curve[i].symbol,
			   0.5 * charwidth, 0.5 * charheight);
		}
	    }
	    ++n;		/* move to next legend position */
	}
    }   /* end legend printing */

    /* calculate the extent of the axes */

    /*-------------------------*/
    /* adjust frame for labels */
    /*-------------------------*/

    /* x-label */

    if (strlen(ez.c_xlabel) > 0)
	ya_min += 3.0 * charheight;
    xlbl_row = xp_min;		/* put x-label on bottom of plot frame */

    /* y-label */

    if (strlen(ez.c_ylabel) > 0)
	xa_min += 3.0 * charwidth;	/* reverse charsize because writing */
    /* text sideways */
    ylbl_col = xp_min + 2 * charwidth;

    /*------------------------------*/
    /* adjust frame for tick labels */
    /*------------------------------*/

    /* calc offset of tick labels from axes */

    if (ez.o_xaxis == NOAXIS || ez.o_xtlabel == FALSE)
	xtl_ofs = 0.0;
    else if (ez.o_xticks == BELOW)
	xtl_ofs = -2.5 * charheight;
    else if (ez.o_xticks == ABOVE)
	xtl_ofs = 1.5 * charheight;

    if (ez.o_yaxis == NOAXIS || ez.o_ytlabel == FALSE)
	ytl_ofs = 0.0;
    else if (ez.o_yticks == LEFT)
	ytl_ofs = -(1 + y_fldwid) * charwidth;
    else if (ez.o_yticks == RIGHT)
	ytl_ofs = 1.5 * charwidth;

    /* see if need to shrink axis extents and/or tick extents */

    if (xtl_ofs != 0.0 && ez.s_ycross == FALSE) {
	if (ez.o_xticks == BELOW) {
	    ya_min += 2.5 * charheight;
	    yt_min = ya_min;
	} else if (ez.o_xticks == ABOVE) {
	    ya_min += 0.0;
	    yt_min = ya_min + 2.5 * charheight;
	}
    } else			/* noaxis, no t-labels, or user set cross */
	yt_min = ya_min;

    if (ytl_ofs != 0.0 && ez.s_xcross == FALSE) {
	if (ez.o_yticks == LEFT) {
	    xa_min += (1 + y_fldwid) * charwidth;
	    xt_min = xa_min;
	} else if (ez.o_yticks == RIGHT) {
	    xa_min += 0.0;
	    xt_min = xa_min + ytl_ofs + y_fldwid * charwidth;
	}
    } else
	xt_min = xa_min;

    xt_max = xa_max;
    yt_max = ya_max;

    /* decrease size of graph, if necessary, to accommadate space */
    /* between axis boundary and boundary of ticks */

    x_added_ticks = -1;
    y_added_ticks = -1;

    if (ez.o_xaxis == NOAXIS || ez.o_xtlabel == FALSE)
	x_added_ticks = 0;
    if (ez.o_yaxis == NOAXIS || ez.o_ytlabel == FALSE)
	y_added_ticks = 0;

    if (ez.o_grid == TRUE) {
	if (x_added_ticks < 0)
	    x_added_ticks = 2;
	if (y_added_ticks < 0)
	    y_added_ticks = 2;
    }

    if (x_added_ticks < 0)
	{
	    if (ez.o_yticks == LEFT || ez.s_ycross)
		x_added_ticks = 1;
	    else
		x_added_ticks = 2;
	}

    if (y_added_ticks < 0)
	{
	    if (ez.o_xticks == BELOW || ez.s_xcross)
		y_added_ticks = 1;
	    else
		y_added_ticks = 2;
	}

    xn_tickinc = (xt_max - xt_min) / (x_nint + x_added_ticks);
    yn_tickinc = (yt_max - yt_min) / (y_nint + y_added_ticks);

    xt_min += 0.5 * x_added_ticks * xn_tickinc;
    xt_max -= 0.5 * x_added_ticks * xn_tickinc;
    yt_min += 0.5 * y_added_ticks * yn_tickinc;
    yt_max -= 0.5 * y_added_ticks * yn_tickinc;

    xgn_min = xt_min;
    xgn_max = xt_max;
    ygn_min = yt_min;
    ygn_max = yt_max;

    /*---------------------------------------------------------------------------*/

    /* PLOT CURVES */

    sgp2_line_style (LS_SOLID);
    drawaxis();

    /* Convert WC in graph boundary to axis boundary */
    {
	double xmin, xmax, ymin, ymax;

	sgp2_window  (xgw_min, ygw_min, xgw_max, ygw_max);	/* Graph boundary */
	sgp2_viewport (xgn_min, ygn_min, xgn_max, ygn_max);

	ndc_to_wc (xa_min, ya_min, &xmin, &ymin);	/* calc WC of axis */
	ndc_to_wc (xa_max, ya_max, &xmax, &ymax);	/* boundaries */

	sgp2_window  (xmin, ymin, xmax, ymax);		/* Set window to axis */
	sgp2_viewport (xa_min, ya_min, xa_max, ya_max);	/* boundaries */
    }
 
    symwidth = charwidth * (xgw_max - xgw_min);
    symheight = charheight * (ygw_max - ygw_min);

    for (ic = 0; ic < ez.i_numcurves; ic++) {
	sgp2_color (ez.curve[ic].color);
	if (ez.curve[ic].linestyle != LS_NOLINE) {
	    sgp2_line_style (ez.curve[ic].linestyle);
	    sgp2_polyline_abs (ez.curve[ic].x, ez.curve[ic].y, ez.curve[ic].numpts);
	}
	if (ez.curve[ic].symbol > 0) {
	    sgp2_line_style(LS_SOLID);
	    sgp2_move_abs (ez.curve[ic].x[0], ez.curve[ic].y[0]);
	    symbol (ez.curve[ic].symbol, symwidth, symheight);
	    for (i = 1; i < ez.curve[ic].numpts; i++)
		if (i % ez.curve[ic].symfreq == 0 || i == ez.curve[ic].numpts - 1) {
		    sgp2_move_abs (ez.curve[ic].x[i], ez.curve[ic].y[i]);
		    symbol (ez.curve[ic].symbol, symwidth, symheight);
		}
	}
    }
 
    if (ez.o_ufinish == FALSE)
	termgrf2 ();
 
    return (TRUE);
}


/* NAME
 *   drawaxis			INTERNAL routine to draw axis & label them
 *
 * SYNOPSIS
 *   drawaxis()
 */

static void drawaxis(void)
{
    double xticklen = 0, yticklen = 0; /* length of ticks in NDC */
    double minorinc;		/* increment between minor axes */
    double xaxispos, yaxispos;	/* crossing of axes */
    double x, y, x2, y2;
    bool axis_near;			/* TRUE if axis too close to print t-label */
    int i, j;
    char str[100];
    char *numstr;

    charsize (charwidth, charheight);
    settextclr (1, -1);

    if (ez.o_xticks == ABOVE)
	xticklen = charheight;
    else if (ez.o_xticks == BELOW)
	xticklen = -charheight;

    if (ez.o_yticks == RIGHT)
	yticklen = charwidth;
    else if (ez.o_yticks == LEFT)
	yticklen = -charwidth;

    sgp2_window  (xp_min, yp_min, xp_max, yp_max);
    sgp2_viewport (xp_min, yp_min, xp_max, yp_max);

    if (strlen (ez.c_title) != 0) {
	sgp2_move_abs (xa_min + (xa_max-xa_min)/2 - strlen(ez.c_title)*charwidth, title_row);
	charsize (charwidth * 2.0, charheight * 2.0);
	settextclr (ez.clr_title, -1);
	sgp2_draw_text (ez.c_title);
	charsize (charwidth, charheight);
    }

    if (ez.o_grid == TRUE || ez.o_box == TRUE) {
	sgp2_color (ez.clr_axis);
	sgp2_move_abs (xa_min, ya_min);
	sgp2_line_abs (xa_max, ya_min);
	sgp2_line_abs (xa_max, ya_max);
	sgp2_line_abs (xa_min, ya_max);
	sgp2_line_abs (xa_min, ya_min);
    }

    /* calculate position of axes */

    /* x-axis */
    if (ez.s_ycross == TRUE) {	/* convert users' world-coord */
	xaxispos = ez.v_ycross;	/* axis to its position in NDC */
	sgp2_window (xgw_min, ygw_min, xgw_max, ygw_max);
	sgp2_viewport (xgn_min, ygn_min, xgn_max, ygn_max);
	x = xgw_min;
	wc_to_ndc (x, xaxispos, &x, &xaxispos);
    } else
	xaxispos = ya_min;

    /* y-axis */
    if (ez.s_xcross == TRUE) {	/* convert users' world-coord */
	yaxispos = ez.v_xcross;	/* axis to its NDC position */
	sgp2_window (xgw_min, ygw_min, xgw_max, ygw_max);
	sgp2_viewport (xgn_min, ygn_min, xgn_max, ygn_max);
	y = ygw_min;
	wc_to_ndc (yaxispos, y, &yaxispos, &y);
    } else
	yaxispos = xa_min;

    /*-------------*/
    /* draw x-axis */
    /*-------------*/

    if (ez.o_xaxis == LINEAR) {
	sgp2_window (xp_min, yp_min, xp_max, yp_max);
	sgp2_viewport (xp_min, yp_min, xp_max, yp_max);

	/* draw axis line */

	sgp2_color (ez.clr_axis);
	if (ez.o_tag && !ez.o_grid && !ez.o_box && ez.s_xcross) {
	    sgp2_move_abs (xa_min, xaxispos - charheight);
	    sgp2_line_abs (xa_min, xaxispos + charheight);
	}
	sgp2_move_abs (xa_min, xaxispos);
	sgp2_line_abs (xa_max, xaxispos);
	if (ez.o_tag && !ez.o_grid && !ez.o_box) {
	    sgp2_move_abs (xa_max, xaxispos - charheight);
	    sgp2_line_abs (xa_max, xaxispos + charheight);
	}

	if (ez.o_grid == TRUE) {
	    sgp2_color (ez.clr_grid);
	    for (i = 0; i <= x_nint; i++) {
		sgp2_move_abs (xt_min + xn_tickinc * i, ya_max);
		sgp2_line_abs (xt_min + xn_tickinc * i, ya_min);
	    }
	}
	sgp2_move_abs (xa_min + (xa_max-xa_min)/2 - strlen(ez.c_xlabel)*charwidth, xlbl_row);
	charsize (charwidth * 2.0, charheight * 2.0);
	settextclr (ez.clr_label, -1);
	sgp2_draw_text (ez.c_xlabel);
	charsize (charwidth, charheight);
	minorinc = xn_tickinc / (ez.o_xminortick + 1);

	for (i = 0; i <= x_nint; i++) {
	    x = xt_min + xn_tickinc * i;
	    sgp2_color (ez.clr_axis);
	    sgp2_move_abs (x, xaxispos);
	    sgp2_line_abs (x, xaxispos + xticklen);
	    if (i != x_nint)
		for (j = 1; j <= ez.o_xminortick; j++) {
		    x2 = x + minorinc * j;
		    sgp2_move_abs (x2, xaxispos);
		    sgp2_line_abs (x2, xaxispos + TICKRATIO * xticklen);
		}
	    axis_near = FALSE;
	    if (xaxispos + xtl_ofs > ya_min && ez.o_yaxis != NOAXIS) {
		double xw, x, y, d;

		xw = xgw_min + i * xw_tickinc;
		sgp2_window (xgw_min, ygw_min, xgw_max, ygw_max);
		sgp2_viewport (xgn_min, ygn_min, xgn_max, ygn_max);
		wc_to_ndc (xw, y, &x, &y);
		sgp2_window (xp_min, yp_min, xp_max, yp_max);
		sgp2_viewport (xp_min, yp_min, xp_max, yp_max);
		d = x - yaxispos;
		if (ez.o_yticks == RIGHT && d >= 0  && d < 0.9 * xn_tickinc)
		    axis_near = TRUE;
		if (ez.o_yticks == LEFT && d <= 0 && d > -0.9 * xn_tickinc)
		    axis_near = TRUE;
	    }

	    if (ez.o_xtlabel == TRUE && axis_near == FALSE) {
		snprintf (str, sizeof(str), x_numfmt, xgw_min + xw_tickinc * i);
		numstr = str_skip_head (str, " ");
		sgp2_move_abs (x-strlen(numstr)*charwidth/2,xaxispos + xtl_ofs);
		settextclr (ez.clr_number, -1);
		sgp2_draw_text (numstr);
	    }
	}
    }		/* x - axis */


    /*--------*/
    /* y-axis */
    /*--------*/

    if (ez.o_yaxis == LINEAR) {
	sgp2_window  (xp_min, yp_min, xp_max, yp_max);
	sgp2_viewport (xp_min, yp_min, xp_max, yp_max);

	sgp2_color (ez.clr_axis);
	if (ez.o_tag && !ez.o_grid && !ez.o_box && ez.s_ycross) {
	    sgp2_move_abs (yaxispos - charwidth, ya_min);
	    sgp2_line_abs (yaxispos + charwidth, ya_min);
	}
	sgp2_move_abs (yaxispos, ya_min);
	sgp2_line_abs (yaxispos, ya_max);
	if (ez.o_tag && !ez.o_grid && !ez.o_box) {
	    sgp2_move_abs (yaxispos - charwidth, ya_max);
	    sgp2_line_abs (yaxispos + charwidth, ya_max);
	}

	if (ez.o_grid == TRUE) {
	    sgp2_color (ez.clr_grid);
	    for (i = 0; i <= y_nint; i++) {
		y = yt_min + yn_tickinc * i;
		sgp2_move_abs (xa_max, y);
		sgp2_line_abs (xa_min, y);
	    }
	}
	sgp2_move_abs (ylbl_col,ya_min + (ya_max-ya_min)/2 - strlen(ez.c_ylabel)*charheight);
	textangle (HALFPI);
	charsize (2 * charheight, 2 * charwidth);	/* axis reversed */
	settextclr (ez.clr_label, -1);
	sgp2_draw_text (ez.c_ylabel);
	textangle (0.0);
	charsize (charwidth, charheight);
	minorinc = yn_tickinc / (ez.o_yminortick + 1);

	for (i = 0; i <= y_nint; i++) {
	    y = yt_min + yn_tickinc * i;
	    sgp2_color (ez.clr_axis);
	    sgp2_move_abs (yaxispos, y);
	    sgp2_line_abs (yaxispos + yticklen, y);
	    if (i != y_nint)
		for (j = 1; j <= ez.o_yminortick; j++) {
		    y2 = y + minorinc * j;
		    sgp2_move_abs (yaxispos, y2);
		    sgp2_line_abs (yaxispos + TICKRATIO * yticklen, y2);
		}
	    axis_near = FALSE;
	    if (yaxispos + ytl_ofs > xa_min && ez.o_xaxis != NOAXIS) {
		double yw, x, y, d;

		yw = ygw_min + i * yw_tickinc;
		sgp2_window (xgw_min, ygw_min, xgw_max, ygw_max);
		sgp2_viewport (xgn_min, ygn_min, xgn_max, ygn_max);
		wc_to_ndc (x, yw, &x, &y);
		sgp2_window (xp_min, yp_min, xp_max, yp_max);
		sgp2_viewport (xp_min, yp_min, xp_max, yp_max);
		d = y - xaxispos;
		if (ez.o_xticks == ABOVE && d >= 0 && d < 0.9 * yn_tickinc)
		    axis_near = TRUE;
		if (ez.o_xticks == BELOW && d <= 0 && d > -0.9 * yn_tickinc)
		    axis_near = TRUE;
	    }
	    if (ez.o_ytlabel == TRUE && axis_near == FALSE) {
		snprintf (str, sizeof(str), y_numfmt, ygw_min + yw_tickinc * i);
		sgp2_move_abs (yaxispos + ytl_ofs, y - 0.5 * charheight);
		settextclr (ez.clr_number, -1);
		sgp2_draw_text (str);
	    }
	}
    }		/* y - axis */
}


static void 
symbol (int sym, double symwidth, double symheight)
{
    if (sym <= 0)
	return;

    if (sym == SB_CROSS) {
	sgp2_move_rel (-0.5 * symwidth, -0.5 * symheight);
	sgp2_line_rel (symwidth, symheight);
	sgp2_move_rel (-symwidth, 0.0);
	sgp2_line_rel (symwidth, -symheight);
	sgp2_move_rel (-0.5 * symwidth, 0.5 * symheight);
    } else if (sym == SB_PLUS) {
	sgp2_move_rel (-0.5 * symwidth, 0.0);
	sgp2_line_rel (symwidth, 0.0);
	sgp2_move_rel (-0.5 * symwidth, -0.5 * symheight);
	sgp2_line_rel (0.0, symheight);
	sgp2_move_rel (0.0, -0.5 * symheight);
    } else if (sym == SB_BOX) {
	sgp2_move_rel (-0.5 * symwidth, -0.5 * symheight);
	sgp2_line_rel (symwidth, 0.0);
	sgp2_line_rel (0.0, symheight);
	sgp2_line_rel (-symwidth, 0.0);
	sgp2_line_rel (0.0, -symheight);
	sgp2_move_rel (0.5 * symwidth, 0.5 * symheight);
    } else if (sym == SB_CIRCLE) {
	sgp2_draw_circle (symwidth);
    } else if (sym == SB_ERRORBAR) {
	sgp2_move_rel (-0.5 * symwidth, 0.5 * symheight);
	sgp2_line_rel (symwidth, 0.0);
	sgp2_move_rel (-0.5 * symwidth, 0.0);
	sgp2_line_rel (0.0, -symheight);
	sgp2_move_rel (-0.5 * symwidth, 0.0);
	sgp2_line_rel (symwidth, 0.0);
	sgp2_move_rel (-0.5 * symwidth, 0.5 * symheight);
    }
}


void ezinit(void)
{
    /* EZPLOT Variables */

    charwidth = DEF_CHARWIDTH;	/* KR_FIX: Make these ez.o_ variables */
    charheight = DEF_CHARHEIGHT;

    /* EZPLOT & EZSET Variables */

    strcpy (ez.c_xlabel, "X axis");
    strcpy (ez.c_ylabel, "Y axis");
    strcpy (ez.c_title, "");
    strcpy (ez.c_legend, "");

    ezplot_firstcall = FALSE;

    ez.i_numcurves = 0;
    ez.i_plotimmediate = FALSE;

    ez.o_reqcurves = 1;
    ez.o_unknowncurves = FALSE;

    ez.o_ustart = FALSE;
    ez.o_ufinish = FALSE;

    ez.o_xporigin = 0.0;
    ez.o_yporigin = 0.0;
    ez.o_xlength  = 1.0;
    ez.o_ylength  = 1.0;

    ez.o_xaxis = LINEAR;
    ez.o_yaxis = LINEAR;

    ez.o_grid = FALSE;
    ez.o_box = FALSE;

    ez.o_xmajortick = 10;
    ez.o_ymajortick =  8;
    ez.o_xminortick =  4;
    ez.o_yminortick =  4;

    ez.o_color = DEF_CURVE_CLR;
    ez.o_symfreq = 1;
    ez.o_symbol = -1;
    ez.o_linestyle = LS_SOLID;

    ez.o_xtlabel = TRUE;
    ez.o_ytlabel = TRUE;
    ez.o_xticks = BELOW;
    ez.o_yticks = LEFT;

    ez.o_legendbox = INSIDE;
    ez.o_tag = FALSE;

    ez.s_xtitle   = FALSE;
    ez.s_ytitle   = FALSE;
    ez.s_xcross   = FALSE;
    ez.s_ycross   = FALSE;
    ez.s_lxfrac   = FALSE;
    ez.s_lyfrac   = FALSE;
    ez.s_xlegend  = FALSE;
    ez.s_ylegend  = FALSE;
    ez.s_textsize = FALSE;

    ez.d_usecrt = TRUE;
    ez.d_useprt = FALSE;
    ez.d_crtmode = DEF_CRTMODE;
    ez.d_prtmode = PRTMODE_DEF;
    ez.d_xprtbuf = XBUF_DEF;
    ez.d_yprtbuf = YBUF_DEF;

    ez.clr_axis   = C_WHITE;		/* set fixed colors */
    ez.clr_title  = (C_CYAN+8);
    ez.clr_label  = (C_CYAN+8);
    ez.clr_legend = (C_RED+8);
    ez.clr_number = (C_GREEN+8);
    ez.clr_grid   = (C_BLACK+8);
}

void ezfree(void)
{
    static char errmsg[] = "EZFREE in EZPLOT";
    int i;
    
    for (i = 0; i < ez.i_numcurves; i++) {
	delete ez.curve[i].x;
	delete ez.curve[i].y;
    }
    ez.i_numcurves = 0;
}

void ezclear(void)
{
    ezfree();
    ezinit();
}
