/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:	   backprojectors.cpp         Classes for backprojection
**   Programmer:   Kevin Rosenberg
**   Date Started: June 2000
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: backprojectors.cpp,v 1.1 2000/06/19 02:59:34 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"


// FUNCTION IDENTIFICATION
//     Backproject* projector = selectBackprojector (...)
//
// PURPOSE
//     Selects a backprojector based on BackprojType 
//     and initializes the backprojector

Backproject* selectBackprojector (BackprojType bjType, const Projections& proj, ImageFile& im, InterpolationType interpType)
{
    Backproject* bj = NULL;

    if (bjType == O_BPROJ_TRIG)
	bj = static_cast<Backproject*>(new BackprojectTrig (proj, im, interpType));
    else if (bjType == O_BPROJ_TABLE)
	bj = static_cast<Backproject*>(new BackprojectTable (proj, im, interpType));
    else if (bjType == O_BPROJ_DIFF)
	bj = static_cast<Backproject*>(new BackprojectDiff (proj, im, interpType));
    else if (bjType == O_BPROJ_DIFF2)
	bj = static_cast<Backproject*>(new BackprojectDiff2 (proj, im, interpType));
    else if (bjType == O_BPROJ_IDIFF2)
	bj = static_cast<Backproject*>(new BackprojectIntDiff2 (proj, im, interpType));
    else 
      sys_error (ERR_WARNING, "Illegal backproject type %d [selectBackprojector]");

    return (bj);
}


// CLASS IDENTICATION
//   Backproject
//
// PURPOSE
//   Pure virtual base class for all backprojectors.

Backproject::Backproject (const Projections& proj, ImageFile& im, const InterpolationType interpType)
  : proj(proj), im(im), interpType(interpType)
{
  detInc = proj.detInc();
  nDet = proj.nDet();
  iDetCenter = (nDet - 1) / 2;	// index refering to L=0 projection 
  rotInc = proj.rotInc();

  v = im.getArray();
  nx = im.nx();
  ny = im.ny();
  im.arrayDataClear();

  xMin = -proj.phmLen() / 2;      // Retangular coords of phantom
  xMax = xMin + proj.phmLen();
  yMin = -proj.phmLen() / 2;
  yMax = yMin + proj.phmLen();

  xInc = (xMax - xMin) / nx;	// size of cells
  yInc = (yMax - yMin) / ny;

  if (interpType != I_NEAREST && interpType != I_LINEAR)
    sys_error (ERR_WARNING, "Illegal interpType %d [selectBackprojector]", interpType);
}

Backproject::~Backproject (void)
{}

void
Backproject::ScaleImageByRotIncrement (void)
{
  for (int ix = 0; ix < nx; ix++)
    for (int iy = 0; iy < ny; iy++)
      v[ix][iy] *= rotInc;
}

void Backproject::errorIndexOutsideDetector (int ix, int iy, double theta, double r, double phi, double L, int iDetPos)
{
    printf ("r=%f, phi=%f\n", r, phi);
    errorIndexOutsideDetector (ix, iy, theta, L, iDetPos);
}

void Backproject::errorIndexOutsideDetector (int ix, int iy, double theta, double L, int iDetPos)
{
    printf ("ix=%d, iy=%d\n", ix, iy);
    printf ("theta=%f, L=%f, detInc=%f\n", theta, L, detInc);
    printf ("proj.ndet=%d, proj.detInc=%.4f, iDetCenter=%d\n", nDet, detInc, iDetCenter);
    printf ("xMin=%15.8f, xMax=%15.8f, xInc=%15.8f\n", xMin, xMax, xInc);
    printf ("yMin=%15.8f, yMax=%15.8f, yInc=%15.8f\n", yMin, yMax, yInc);
    sys_error (ERR_WARNING, "iDetPos index outside bounds: %d [backprojector]", iDetPos);
}


// CLASS IDENTICATION
//   BackprojectTrig
//
// PURPOSE
//   Uses trigometric functions at each point in image for backprojection.

void
BackprojectTrig::BackprojectView (const double* const filteredProj, const double view_angle)
{
  double theta = HALFPI + view_angle;   // Add PI/2 to get perpendicular angle to detector	
  int ix, iy;
  double x, y;			// Rectang coords of center of pixel 

  for (x = xMin + xInc / 2, ix = 0; ix < nx; x += xInc, ix++)
    for (y = yMin + yInc / 2, iy = 0; iy < ny; y += yInc, iy++) {
      double r = sqrt (x * x + y * y);   // distance of cell from center
      double phi = atan2 (y, x);         // angle of cell from center
      double L = r * cos (theta - phi);  // position on detector

      if (interpType == I_NEAREST) {
	int iDetPos = iDetCenter + nearest<int> (L / detInc); // calc'd index in the filter raysum array

	if (iDetPos < 0 || iDetPos >= nDet)	// check for impossible: index outside of raysum pos 
	    errorIndexOutsideDetector (ix, iy, theta, r, phi, L, iDetPos);
	else
	  v[ix][iy] += rotInc * filteredProj[iDetPos];
      } else if (interpType == I_LINEAR) {
	  double p = L / detInc;	// position along detector
	  double pFloor = floor (p);
	  int iDetPos = iDetCenter + static_cast<int>(pFloor);
	  double frac = p - pFloor;	// fraction distance from det
	  if (iDetPos < 0 || iDetPos >= nDet - 1) 	// check for impossible: index outside of raysum pos 
	    errorIndexOutsideDetector (ix, iy, theta, r, phi, L, iDetPos);
	  else
	    v[ix][iy] += rotInc * ((1-frac) * filteredProj[iDetPos] + frac * filteredProj[iDetPos+1]);
      }
    }
}  


// CLASS IDENTICATION
//   BackprojectTable
//
// PURPOSE
//   Precalculates trigometric function value for each point in image for backprojection.

BackprojectTable::BackprojectTable (const Projections& proj, ImageFile& im, InterpolationType interpType)
  : Backproject::Backproject (proj, im, interpType)
{
  arrayR.initSetSize (nx, ny);
  arrayPhi.initSetSize (nx, ny);
  r = arrayR.getArray();
  phi = arrayPhi.getArray();

  double x, y;			// Rectang coords of center of pixel 
  int ix, iy;
  for (x = xMin + xInc / 2, ix = 0; ix < nx; x += xInc, ix++)
    for (y = yMin + yInc / 2, iy = 0; iy < ny; y += yInc, iy++) {
      r[ix][iy] = sqrt (x * x + y * y);
      phi[ix][iy] = atan2 (y, x);
    }
}

BackprojectTable::~BackprojectTable (void)
{
  ScaleImageByRotIncrement();
}

void
BackprojectTable::BackprojectView (const double* const filteredProj, const double view_angle)
{
  double theta = HALFPI + view_angle;  // add half PI to view angle to get perpendicular theta angle

  for (int ix = 0; ix < nx; ix++) {
    ImageFileColumn pImCol = v[ix];

    for (int iy = 0; iy < ny; iy++) {
      double L = r[ix][iy] * cos (theta - phi[ix][iy]);

      if (interpType == I_NEAREST) {
	int iDetPos = iDetCenter + nearest<int>(L / detInc);	// calc index in the filtered raysum vector 

	if (iDetPos < 0 || iDetPos >= nDet) 	// check for impossible: index outside of raysum pos 
	  errorIndexOutsideDetector (ix, iy, theta, r[ix][iy], phi[ix][iy], L, iDetPos);
	else
	  pImCol[iy] += filteredProj[iDetPos];
      } else if (interpType == I_LINEAR) {
	double dPos = L / detInc;		// position along detector 
	double dPosFloor = floor (dPos);
	int iDetPos = iDetCenter + static_cast<int>(dPosFloor);
	double frac = dPos - dPosFloor;	// fraction distance from det 
	if (iDetPos < 0 || iDetPos >= nDet - 1)
	    errorIndexOutsideDetector (ix, iy, theta, r[ix][iy], phi[ix][iy], L, iDetPos);
	else
	  pImCol[iy] += ((1-frac) * filteredProj[iDetPos] + frac * filteredProj[iDetPos+1]);
      }
    }	// end for y 
  }	// end for x 
}


// CLASS IDENTICATION
//   BackprojectDiff
//
// PURPOSE
//   Backprojects by precalculating the change in L position for each x & y step in the image.
//   Iterates in x & y direction by adding difference in L position

BackprojectDiff::BackprojectDiff (const Projections& proj, ImageFile& im, InterpolationType interpType)
  :  Backproject::Backproject (proj, im, interpType)
{
  // calculate center of first pixel v[0][0] 
  double x = xMin + xInc / 2;
  double y = yMin + yInc / 2;
  start_r = sqrt (x * x + y * y);
  start_phi = atan2 (y, x);

  im.arrayDataClear();
}

BackprojectDiff::~BackprojectDiff()
{
  ScaleImageByRotIncrement();
}

void
BackprojectDiff::BackprojectView (const double* const filteredProj, const double view_angle)
{
  double theta = - view_angle;  // add half PI to view angle to get perpendicular theta angle
  double det_dx = xInc * sin (theta);
  double det_dy = yInc * cos (theta);
  double lColStart = start_r * cos (theta - start_phi);  // calculate L for first point in image
	
  for (int ix = 0; ix < nx; ix++, lColStart += det_dx) {
    double curDetPos = lColStart;
    ImageFileColumn pImCol = v[ix];
  
    for (int iy = 0; iy < ny; iy++, curDetPos += det_dy) {
#ifdef DEBUG
      printf ("[%2d,%2d]:  %8.5lf  ", ix, iy, curDetPos);
#endif
      if (interpType == I_NEAREST) {
	int iDetPos = iDetCenter + nearest<int>(curDetPos / detInc);	// calc index in the filtered raysum vector 

	if (iDetPos < 0 || iDetPos >= nDet)	// check for impossible: index outside of raysum pos 
	    errorIndexOutsideDetector (ix, iy, theta, curDetPos, iDetPos);
	else
	  pImCol[iy] += filteredProj[iDetPos];
      } else if (interpType == I_LINEAR) {
	double detPos = curDetPos / detInc;		// position along detector 
	double detPosFloor = floor (detPos);
	int iDetPos = iDetCenter + static_cast<int>(detPosFloor);
	double frac = detPos - detPosFloor;	// fraction distance from det 
	if (iDetPos < 0 || iDetPos >= nDet - 1)
	    errorIndexOutsideDetector (ix, iy, theta, curDetPos, iDetPos);
	else
	  pImCol[iy] += ((1-frac) * filteredProj[iDetPos] + frac * filteredProj[iDetPos+1]);
      }
    }	// end for y 
  }	// end for x 
}


// CLASS IDENTICATION
//   BackprojectDiff2
//
// PURPOSE
//   Optimized version of BackprojectDiff

void
BackprojectDiff2::BackprojectView (const double* const filteredProj, const double view_angle)
{
  double theta = - view_angle;  // add half PI to view angle to get perpendicular theta angle

  // Distance betw. detectors for an angle given in units of detectors 
  double det_dx = xInc * sin (theta) / detInc;
  double det_dy = yInc * cos (theta) / detInc;

  // calculate detPosition for first point in image (ix=0, iy=0) 
  double detPosColStart = start_r * cos (theta - start_phi) / detInc;
	
#ifdef DEBUG
  printf ("start_r=%8.5f, start_phi=%8.5f, rotInc=%8.5f\n", start_r, start_phi, rotInc);
#endif
  for (int ix = 0; ix < nx; ix++, detPosColStart += det_dx) {
    double curDetPos = detPosColStart;
    ImageFileColumn pImCol = v[ix];

    for (int iy = 0; iy < ny; iy++, curDetPos += det_dy) {
#ifdef DEBUG
      printf ("[%2d,%2d]: %8.5f %8.5f\n", ix, iy, curDetPos, filteredProj[iDetCenter + nearest<int>(L))]);
#endif
      if (interpType == I_NEAREST) {
	int iDetPos = iDetCenter + nearest<int> (curDetPos);	// calc index in the filtered raysum vector 
	
	if (iDetPos < 0 || iDetPos >= nDet) 	// check for impossible: index outside of raysum pos 
	    errorIndexOutsideDetector (ix, iy, theta, curDetPos, iDetPos);
	else
	  *pImCol++ += filteredProj[iDetPos];
      } else if (interpType == I_LINEAR) {
	double detPosFloor = floor (curDetPos);
	int iDetPos = iDetCenter + static_cast<int>(detPosFloor);
	double frac = curDetPos - detPosFloor;	// fraction distance from det 
	if (iDetPos < 0 || iDetPos >= nDet - 1)
	    errorIndexOutsideDetector (ix, iy, theta, curDetPos, iDetPos);
	else
	  *pImCol++ += ((1-frac) * filteredProj[iDetPos] + frac * filteredProj[iDetPos+1]);
      }
    }	// end for y
  }	// end for x
}

// CLASS IDENTICATION
//   BackprojectIntDiff2
//
// PURPOSE
//   Integer version of BackprojectDiff2

void
BackprojectIntDiff2::BackprojectView (const double* const filteredProj, const double view_angle)
{
  double theta = - view_angle;  // add half PI to view angle to get perpendicular theta angle

#if SIZEOF_LONG == 8
  long int scale = 1 << 32;
#else
  long int scale = 1 << 16;
#endif
  double dScale = scale;
  long int halfScale = scale / 2;

  long int det_dx = nearest<long int> (xInc * sin (theta) / detInc * scale);
  long int det_dy = nearest<long int> (yInc * cos (theta) / detInc * scale);

  // calculate L for first point in image (0, 0) 
  long int detPosColStart = nearest<long int> (start_r * cos (theta - start_phi) / detInc * scale);
	
  for (int ix = 0; ix < nx; ix++, detPosColStart += det_dx) {
    long int curDetPos = detPosColStart;
    ImageFileColumn pImCol = v[ix];

    for (int iy = 0; iy < ny; iy++, curDetPos += det_dy) {
      if (interpType == I_NEAREST) {
	int detPosNearest = (curDetPos >= 0 ? ((curDetPos + halfScale) / scale) : ((curDetPos - halfScale) / scale));
	int iDetPos = iDetCenter + detPosNearest;	// calc index in the filtered raysum vector 

	if (iDetPos < 0 || iDetPos >= nDet)  // check for impossible: index outside of raysum pos 
	    errorIndexOutsideDetector (ix, iy, theta, curDetPos, iDetPos);
	else
	  *pImCol++ += filteredProj[iDetPos];
      } else if (interpType == I_LINEAR) {
	long int detPosFloor = curDetPos / scale;
	long int detPosRemainder = curDetPos % scale;
	if (detPosRemainder < 0) {
	  detPosFloor--;
	  detPosRemainder += scale;
	}
	int iDetPos = iDetCenter + detPosFloor;
	double frac = detPosRemainder / dScale;
	if (iDetPos < 0 || iDetPos >= nDet - 1)
	    errorIndexOutsideDetector (ix, iy, theta, curDetPos, iDetPos);
	else
	  *pImCol++ += ((1-frac) * filteredProj[iDetPos] + frac * filteredProj[iDetPos+1]);
      }
    }	// end for y
  }	// end for x
}
