/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:	    scanner.cpp
**   Purpose:       Classes for CT scanner
**   Programmer:    Kevin Rosenberg
**   Date Started:  1984
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: scanner.cpp,v 1.1 2000/06/19 02:59:34 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"


// NAME
//   DetectorArray	 Construct a DetectorArray

DetectorArray::DetectorArray (const int nDet)
{
  m_nDet = nDet;
  m_detValues = new DetectorValue [m_nDet];
}


// NAME
//   ~DetectorArray		Free memory allocated to a detector array

DetectorArray::~DetectorArray (void)
{
  delete [] m_detValues;
}



/* NAME
 *   Scanner::Scanner		Construct a user specified detector structure
 *
 * SYNOPSIS
 *   Scanner (phm, nDet, nView, nSample)
 *   Phantom& phm		PHANTOM that we are making detector for
 *   int geomety                Geometry of detector
 *   int nDet			Number of detector along detector array
 *   int nView			Number of rotated views
 *   int nSample		Number of rays per detector
 */

Scanner::Scanner (const Phantom& phm, const ScannerGeometry geometry, int nDet, int nView, int nSample, const double rot_anglen)
{
  m_phmLen = phm.maxAxisLength();      // maximal length along an axis

  if (nView < 1)
    nView = 1;
  if (nSample < 1)
    m_nSample = 1;
  if (nDet < 1)
    nDet = 1;
  if ((nDet % 2) == 0)
    ++nDet;		// ensure odd number of detectors

  m_geometry = geometry;
  m_nDet     = nDet;
  m_nView    = nView;
  m_nSample  = nSample;
  m_detLen   = SQRT2 * m_phmLen * ((m_nDet + N_EXTRA_DETECTORS) / static_cast<double>(m_nDet));

  m_rotLen   =  rot_anglen;

  m_radius  = m_detLen / 2;
  m_detInc  = m_detLen / m_nDet;
  m_rotInc  = m_rotLen / m_nView;

  m_initPos.xd1 = m_detLen / 2;
  m_initPos.yd1 = -m_detLen / 2;
  m_initPos.xd2 = m_detLen / 2;
  m_initPos.yd2 = m_detLen / 2;
  m_initPos.xs1 = -m_detLen / 2;
  m_initPos.ys1 = -m_detLen / 2;
  m_initPos.xs2 = -m_detLen / 2;
  m_initPos.ys2 = m_detLen / 2;
  m_initPos.angle = 0.0;
}

Scanner::~Scanner (void)
{
}



/* NAME
 *   raysum_collect				Calculate ray sums for a Phantom
 *
 * SYNOPSIS
 *   rs = raysum_collect (det, phm, start_view, trace, unit_pulse)
 *   Scanner& det			Scanner specifications**
 *   RAYSUM *rs				Calculated ray sum matrix
 *   Phantom& phm			Phantom we are taking ray sums of
 *   int trace				Boolean flag to signal ray sum tracing
*/

void
Scanner::collectProjections (Projections& proj, const Phantom& phm, const int start_view, const int trace)
{
  GRFMTX_2D rotmtx_initial, temp;
  GRFMTX_2D rotmtx_incr;

  double start_angle = start_view * proj.rotInc();
  double xcent = phm.xmin() + (phm.xmax() - phm.xmin()) / 2;
  double ycent = phm.ymin() + (phm.ymax() - phm.ymin()) / 2;

  double xd1 = xcent + m_initPos.xd1;
  double yd1 = ycent + m_initPos.yd1;
  double xd2 = xcent + m_initPos.xd2;
  double yd2 = ycent + m_initPos.yd2;
  double xs1 = xcent + m_initPos.xs1;
  double ys1 = ycent + m_initPos.ys1;
  double xs2 = xcent + m_initPos.xs2;
  double ys2 = ycent + m_initPos.ys2;
 
  m_trace = trace;

#ifdef HAVE_SGP 
  SGP_ID gid;
  if (m_trace >= TRACE_PHM) {
    double wsize = 1.42 * m_phmLen / 2;	/* sqrt(2) * radius */
      
    gid = sgp2_init (512, 512, "RayCollect");
    sgp2_color (C_LTBLUE);
    sgp2_window (xcent - wsize, ycent - wsize, xcent + wsize, ycent + wsize);
    sgp2_color (C_BROWN);
#if RADIUS
    sgp2_draw_circle (m_phmLen / 2);
#else
    sgp2_draw_rect (xcent - m_phmLen / 2, ycent - m_phmLen / 2,
		    xcent + m_phmLen / 2, ycent + m_phmLen / 2);
#endif
    sgp2_color (C_BROWN);
    sgp2_move_abs (0., 0.);
    sgp2_draw_circle (wsize);
    //      raysum_trace_menu_column = (crt->xsize * crt->asp) / 8 + 3;
    traceShowParam ("X-Ray Simulator", "%s", RAYSUM_TRACE_ROW_TITLE, 8+C_LTWHITE, " ");
    traceShowParam ("---------------", "%s", RAYSUM_TRACE_ROW_TITLE2, 8+C_LTWHITE, " ");
    traceShowParam ("Phantom:",       "%s", RAYSUM_TRACE_ROW_PHANT_ID, C_YELLOW, " Herman");
    traceShowParam ("Chomaticity  :", "%s", RAYSUM_TRACE_ROW_CHROMATIC, C_LTGREEN, "Mono");
    traceShowParam ("Scatter      :", "%5.1f", RAYSUM_TRACE_ROW_SCATTER, C_LTGREEN, 0.);
    traceShowParam ("Photon Uncert:", "%5.1f", RAYSUM_TRACE_ROW_PHOT_STAT, C_LTGREEN, 0.);
    traceShowParam ("Num Scanners:", "%5d", RAYSUM_TRACE_ROW_NDET, C_LTRED, proj.nDet());
    traceShowParam ("Num Views    :", "%5d", RAYSUM_TRACE_ROW_NVIEW, C_LTRED, proj.nView());
    traceShowParam ("Samples / Ray:", "%5d", RAYSUM_TRACE_ROW_SAMPLES, C_LTRED, m_nSample);
    
    sgp2_color (C_LTGREEN);
    phm.draw();
    
    initmarker (BDIAMOND, 129);
  }
#endif

/* Calculate initial rotation matrix */
  xlat_mtx2 (rotmtx_initial, -xcent, -ycent);
  rot_mtx2 (temp, start_angle);
  mult_mtx2 (rotmtx_initial, temp, rotmtx_initial);
  xlat_mtx2 (temp, xcent, ycent);
  mult_mtx2 (rotmtx_initial, temp, rotmtx_initial);

  xform_mtx2 (rotmtx_initial, xd1, yd1);	/* rotate detector endpoints */
  xform_mtx2 (rotmtx_initial, xd2, yd2);      /* to initial view_angle */
  xform_mtx2 (rotmtx_initial, xs1, ys1);
  xform_mtx2 (rotmtx_initial, xs2, ys2);

/* Calculate incrementatal rotation matrix */
  xlat_mtx2 (rotmtx_incr, -xcent, -ycent);
  rot_mtx2 (temp, proj.rotInc());
  mult_mtx2 (rotmtx_incr, temp, rotmtx_incr);
  xlat_mtx2 (temp, xcent, ycent);
  mult_mtx2 (rotmtx_incr, temp, rotmtx_incr);
  
  int iview;
  double viewAngle;
  for (iview = 0, viewAngle = start_angle;  iview < proj.nView(); iview++, viewAngle += proj.rotInc()) {
      DetectorArray& detArray = proj.getDetectorArray( iview );

#ifdef HAVE_SGP
    if (m_trace >= TRACE_PHM) {
      sgp2_move_abs (xd1, yd1);
      sgp2_line_abs (xd2, yd2);
      sgp2_move_abs (xs1, ys1);
      sgp2_line_abs (xs2, ys2);
    }
#endif
    if (m_trace)
      traceShowParam ("Current View :", "%5d", RAYSUM_TRACE_ROW_CURR_VIEW, C_LTMAGENTA, iview);
	    
    projectSingleView (phm, detArray, xd1, yd1, xd2, yd2, xs1, ys1, xs2, ys2);
    detArray.setViewAngle (viewAngle);
      
#ifdef HAVE_SGP
    if (m_trace >= TRACE_PHM) {
      //	rs_plot (detArray, xd1, yd1, xcent, ycent, theta);
      sgp2_move_abs (xd1, yd1);
      sgp2_line_abs (xd2, yd2);
      sgp2_move_abs (xs1, ys1);
      sgp2_line_abs (xs2, ys2);
    }
#endif
    xform_mtx2 (rotmtx_incr, xd1, yd1);	 // rotate detector endpoints 
    xform_mtx2 (rotmtx_incr, xd2, yd2);
    xform_mtx2 (rotmtx_incr, xs1, ys1);
    xform_mtx2 (rotmtx_incr, xs2, ys2);
  } /* for each iview */
}


/* NAME
 *    rayview			Calculate raysums for a view at any angle
 *
 * SYNOPSIS
 *    rayview (phm, detArray, xd1, nSample, yd1, xd2, yd2, xs1, ys1, xs2, ys2)
 *    Phantom& phm		Phantom to scan
 *    DETARRAY *detArray		Storage of values for detector array
 *    Scanner& det		Scanner parameters
 *    double xd1, yd1, xd2, yd2	Beginning & ending detector positions
 *    double xs1, ys1, xs2, ys2	Beginning & ending source positions
 *
 * RAY POSITIONING
 *         For each detector, have there are a variable number of rays traced.
 *     The source of each ray is the center of the source x-ray cell. The
 *     detector positions are equally spaced within the cell
 *
 *         The increments between rays are calculated so that the cells start
 *     at the beginning of a detector cell and they end on the endpoint
 *     of the cell.  Thus, the last cell starts at (xd2-ddx),(yd2-ddy).
 *         The exception to this is if there is only one ray per detector.
 *     In that case, the detector position is the center of the detector cell.
 */

void 
Scanner::projectSingleView (const Phantom& phm, DetectorArray& detArray, const double xd1, const double yd1, const double xd2, const double yd2, const double xs1, const double ys1, const double xs2, const double ys2)
{
  double ddx = (xd2 - xd1) / detArray.nDet();  // change in coords between detectors
  double ddy = (yd2 - yd1) / detArray.nDet();
  double sdx = (xs2 - xs1) / detArray.nDet();  // change in coords between source
  double sdy = (ys2 - ys1) / detArray.nDet();

  double ddx2 = ddx / m_nSample;	// Incr. between rays with detector cell
  double ddy2 = ddy / m_nSample;	// Doesn't include detector endpoints 
  double ddx2_ofs = ddx2 / 2;           // offset of 1st ray from start of detector cell
  double ddy2_ofs = ddy2 / 2;
  
  double xd_maj = xd1 + ddx2_ofs;       // Incr. between detector cells
  double yd_maj = yd1 + ddy2_ofs;
  double xs_maj = xs1 + (sdx / 2);	// put ray source in center of cell 
  double ys_maj = ys1 + (sdy / 2);

  DetectorValue* detval = detArray.detValues();

  if (phm.getComposition() == P_UNIT_PULSE) {  // put unit pulse in center of view
    for (int d = 0; d < detArray.nDet(); d++)
      if (detArray.nDet() / 2 == d && (d % 2) == 1)
	detval[d] = 1;
      else
	detval[d] = 0;
  } else {
    for (int d = 0; d < detArray.nDet(); d++) {
      double xd = xd_maj;
      double yd = yd_maj;
      double xs = xs_maj;
      double ys = ys_maj;
      double sum = 0.0;
      for (int i = 0; i < m_nSample; i++) {
#ifdef HAVE_SGP
	if (m_trace >= TRACE_RAYS) {
	  sgp2_move_abs (xs, ys);
	  sgp2_line_abs (xd, yd);
	}
#endif
	sum += projectSingleLine (phm, xd, yd, xs, ys);
	      
	if (m_trace >= TRACE_RAYS)
	  traceShowParam ("Attenuation  :", "%5.2f", RAYSUM_TRACE_ROW_ATTEN, C_LTMAGENTA, "sum");

#ifdef HAVE_SGP
	if (m_trace >= TRACE_RAYS) {
	  sgp2_move_abs (xs, ys);
	  sgp2_line_abs (xd, yd);
	}
#endif
	xd += ddx2;
	yd += ddy2;
      }

      detval[d] = sum / m_nSample;
      xd_maj += ddx;
      yd_maj += ddy;
      xs_maj += sdx;
      ys_maj += sdy;
    } /* for each detector */
  } /* if not unit pulse */
}


void 
Scanner::traceShowParam (const char *label, const char *fmt, int row, int color, ...)
{  
  char s[80];
  va_list arg;

  va_start(arg, color);
  //  cio_set_cpos (raysum_trace_menu_column, row);
  snprintf (s, sizeof(s), label, "%s");
  //  cio_set_text_clr (color - 8, 0);
  cio_put_str (s);
  vsnprintf (s, sizeof(s), fmt, arg);
  //  cio_set_text_clr (color, 0);
  cio_put_str (s);

  va_end(arg);
}


/* NAME
 *    projectSingleLine			INTERNAL: Calculates raysum along a line for a Phantom
 *
 * SYNOPSIS
 *    rsum = phm_ray_attenuation (phm, x1, y1, x2, y2)
 *    double rsum		Ray sum of Phantom along given line
 *    Phantom& phm;		Phantom from which to calculate raysum
 *    double *x1, *y1, *x2, y2	Endpoints of ray path (in Phantom coords)
 */

double 
Scanner::projectSingleLine (const Phantom& phm, const double x1, const double y1, const double x2, const double y2)
{
  // check ray against each pelem in Phantom 
  double rsum = 0.0;
  for (PElemConstIterator i = phm.listPElem().begin(); i != phm.listPElem().end(); i++)
    rsum += projectLineAgainstPElem (**i, x1, y1, x2, y2);

  return (rsum);
}


/* NAME
 *   pelem_ray_attenuation		Calculate raysum of an pelem along one line
 *
 * SYNOPSIS
 *   rsum = pelem_ray_attenuation (pelem, x1, y1, x2, y2)
 *   double rsum		Computed raysum
 *   PhantomElement& pelem		Pelem to scan
 *   double x1, y1, x2, y2	Endpoints of raysum line
 */

double 
Scanner::projectLineAgainstPElem (const PhantomElement& pelem, double x1, double y1, double x2, double y2)
{
  if (! pelem.clipLineWorldCoords (x1, y1, x2, y2)) {
    if (m_trace == TRACE_CLIPPING)
      cio_tone (1000., 0.05);
    return (0.0);
  }

#ifdef HAVE_SGP
  if (m_trace == TRACE_CLIPPING) {
    sgp2_move_abs (x1, y1);
    sgp2_line_abs (x2, y2);
    cio_tone (8000., 0.05);
    sgp2_move_abs (x1, y1);
    sgp2_line_abs (x2, y2);
  }
#endif

  double len = lineLength (x1, y1, x2, y2);
  return (len * pelem.atten());
}

